@@ -5,18 +5,16 @@ A Python driver for Alicat mass flow controllers, using serial communication.
 Distributed under the GNU General Public License v2
 Copyright (C) 2015 NuMat Technologies
 """
+import io
 import serial
 
 
-class FlowController(object):
-    """Python driver for [Alicat Flow Controllers](http://www.alicat.com/
-    products/mass-flow-meters-and-controllers/mass-flow-controllers/).
+class FlowMeter(object):
+    """Python driver for [Alicat Flow Meters](http://www.alicat.com/
+    products/mass-flow-meters-and-controllers/mass-flow-meters/).
 
-    This communicates with the flow controller over a USB or RS-232/RS-485
+    This communicates with the flow meter over a USB or RS-232/RS-485
     connection using pyserial.
-
-    To set up your Alicat flow controller, power on the device and make sure
-    that the "Input" option is set to "Serial".
     """
     def __init__(self, port="/dev/ttyUSB0", address="A"):
         """Connects this driver with the appropriate USB / serial port.
@@ -27,6 +25,8 @@ class FlowController(object):
         """
         self.address = address
         self.connection = serial.Serial(port, 19200, timeout=1)
+        bp = io.BufferedRWPair(self.connection, self.connection, 1)
+        self.ser = io.TextIOWrapper(bp, newline="\r", line_buffering=True)
         self.keys = ["pressure", "temperature", "volumetric_flow", "mass_flow",
                      "flow_setpoint", "gas"]
         self.gases = ["Air", "Ar", "CH4", "CO", "CO2", "C2H6", "H2", "He",
@@ -35,7 +35,7 @@ class FlowController(object):
                       "C-8", "C-2", "C-75", "A-75", "A-25", "A1025", "Star29",
                       "P-5"]
 
-    def get(self):
+    def get(self, retries=2):
         """Get the current state of the flow controller.
 
         From the Alicat mass flow controller documentation, this data is:
@@ -43,33 +43,35 @@ class FlowController(object):
          * Temperature (normally in C)
          * Volumetric flow (in units specified at time of order)
          * Mass flow (in units specified at time of order)
-         * Flow setpoint
+         * Flow setpoint (only for flow controllers)
          * Total flow (only on models with the optional totalizer function)
          * Currently selected gas
 
+        Args:
+            retries: Number of times to re-attempt reading. Default 2.
         Returns:
             The state of the flow controller, as a dictionary.
         """
+        self.connection.flush()
         command = "*@={addr}\r\n".format(addr=self.address)
-        self.connection.write(command)
-        line = self.connection.readline().split()
-        address, values = line[0], line[1:]
+        for _ in range(retries+1):
+            self.connection.write(command)
+            line = self.ser.readline()
+            if line:
+                break
+        else:
+            raise IOError("Could not read from flow controller.")
+        spl = line.split()
+        address, values = spl[0], spl[1:]
         if address != self.address:
-            raise Exception("Flow controller address mismatch.")
-        if len(values) == 7 and len(self.keys) == 6:
+            raise ValueError("Flow controller address mismatch.")
+        if len(values) == 5 and len(self.keys) == 6:
+            del self.keys[-2]
+        elif len(values) == 7 and len(self.keys) == 6:
             self.keys.insert(5, "total flow")
         return {k: (v if k == self.keys[-1] else float(v))
                 for k, v in zip(self.keys, values)}
 
-    def set_flow_rate(self, flow):
-        """Sets the target flow rate.
-
-        Args:
-            flow: The target flow rate, in units specified at time of purchase
-        """
-        command = "{addr}S{flow:.2f}\r\n".format(addr=self.address, flow=flow)
-        self.connection.write(command)
-
     def set_gas(self, gas):
         """Sets the gas type.
 
@@ -81,15 +83,47 @@ class FlowController(object):
                 "C-75", "A-75", "A-25", "A1025", "Star29", "P-5"
         """
         if gas not in self.gases:
-            raise Exception("{} not supported!".format(gas))
-        command = "{addr}$${gas}\r\n".format(addr=self.address, gas=gas)
+            raise ValueError("{} not supported!".format(gas))
+        self.connection.flush()
+        command = "{addr}$${gas}\r\n".format(addr=self.address,
+                                             gas=self.gases.index(gas))
         self.connection.write(command)
+        line = self.ser.readline()
+        if not line or line.split()[-1] != gas:
+            raise IOError("Could not set gas type")
 
     def close(self):
         """Closes the serial port. Call this on program termination."""
+        self.connection.flush()
         self.connection.close()
 
 
+class FlowController(FlowMeter):
+    """Python driver for [Alicat Flow Controllers](http://www.alicat.com/
+    products/mass-flow-meters-and-controllers/mass-flow-controllers/).
+
+    This communicates with the flow controller over a USB or RS-232/RS-485
+    connection using pyserial.
+
+    To set up your Alicat flow controller, power on the device and make sure
+    that the "Input" option is set to "Serial".
+    """
+    def set_flow_rate(self, flow):
+        """Sets the target flow rate.
+
+        Args:
+            flow: The target flow rate, in units specified at time of purchase
+        """
+        self.connection.flush()
+        command = "{addr}S{flow:.2f}\r\n".format(addr=self.address, flow=flow)
+        self.connection.write(command)
+        line = self.ser.readline()
+        while line and len(line.split()) > 5:
+            line = self.ser.readline()
+        if not line or abs(float(line) - flow) > 0.01:
+            raise IOError("Could not set flow.")
+
+
 def command_line():
     import argparse
     import json
@@ -121,7 +155,7 @@ def command_line():
     if args.set_gas:
         flow_controller.set_gas(args.set_gas)
     if args.set_flow_rate is not None:
-        flow_controller.set_flow_rate(args.flow_rate)
+        flow_controller.set_flow_rate(args.set_flow_rate)
 
     if args.stream:
         try:
